package com.jh.boot.jpa;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Function;

import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;
import javax.persistence.criteria.CriteriaQuery;

import org.hibernate.query.NativeQuery;


/**
 * TODO comment.
 *
 * @author jh
 */
public abstract class AbstractHibernateRepository {

    /** The session factory. */
    @PersistenceContext
    protected EntityManager entityManager;



    /**
     * Criteria.
     *
     * @param clazz the clazz
     * @return the criteria
     */
    protected <T> CriteriaQuery<T> criteria(Class<T> clazz) {
    	return entityManager.getCriteriaBuilder().createQuery(clazz);
    }
    
    
    @SuppressWarnings("unchecked")
	protected <T> NativeQuery<T> nativeQuery(String query) {
    	return (NativeQuery<T>) entityManager.createNativeQuery(query);
    }
    
    @SuppressWarnings("unchecked")
	protected <T> T singleResult(Query q) {
    	try {
    		return (T) q.getSingleResult();
    	} catch (NoResultException e) {
    		return null;
    	}
    }

    
	public <ID, T> T findById(Class<T> clazz, ID id) {
		return entityManager.find(clazz, id);
	}
	
	
	/**
	 * TODO.
	 *
	 * @param <T> the generic type
	 * @param entity the entity
	 */
    public <T extends AbstractIdEntity> void save(T entity) {
        if (entity.getId() == null) {
            entityManager.persist(entity);
        } else {
            entityManager.merge(entity);
        }
    }

    /**
     * Delete.
     *
     * @param <T> the generic type
     * @param entity the entity
     */
    public <T extends AbstractIdEntity> void delete(T entity) {
        if (entity.getId() != null) {
        	entityManager.remove(entity);
        }
    }

	/**
	 * Sync references.
	 *
	 * @param <T> the generic type
	 * @param <S> the generic type
	 * @param entity the entity
	 * @param referencedClass the referenced class
	 * @param referencedField the referenced field
	 * @param references the references
	 */
	//FIXME??
	protected <T extends AbstractIdEntity, S extends AbstractIdEntity> void syncReferences(T entity, String referencedClass, String referencedField, Collection<S> references) {
		if (entity.getId() == null) {
			//nothing to do
			return;
		}
		
		//filter only already stored references
		Set<S> elements = new HashSet<>();
		if (references != null) {
			references.stream().filter(e -> e.getId() != null).forEach(e -> elements.add(e));
		}
		
		StringBuilder sb = new StringBuilder("delete from ");
		sb.append(referencedClass).append(" ref");
		sb.append(" where ref.").append(referencedField).append(" = :root");
		
		if (!elements.isEmpty()) {
			sb.append(" and ref not in (:refs)");
		}
		
    	Query q = entityManager.createQuery(sb.toString());
    	q.setParameter("root", entity);
    	if (!elements.isEmpty()) {
    		q.setParameter("refs", elements);
    	}

    	q.executeUpdate();
    	if (references != null) {
        	for (S s : references) {
        	    save(s);
        	}
    	}
	}
	
	
	/**
	 * Sync references.
	 *
	 * @param <T> the generic type
	 * @param <S> the generic type
	 * @param entities the entities
	 * @param referencedClass the referenced class
	 * @param referencedField the referenced field
	 * @param valueProviders the value providers
	 */
	protected <T extends AbstractIdEntity, S extends AbstractIdEntity> void syncReferences(Collection<T> entities, String referencedClass, String referencedField, Function<T, ? extends Collection<S>> valueProviders) {
		if (entities == null || entities.isEmpty()) {
			//nothing to do
			return;
		}
		
		for (T entity : entities) {
			if (entity.getId() == null) {
				continue;
			}
			syncReferences(entity, referencedClass, referencedField, valueProviders.apply(entity));
		}
	}

	
}
